#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <limits.h>
#include <math.h>

#define Q_N	16
#define Q_M	16
#define Q_SCALE	(1 << Q_N)	/* Q_SCALE = 2^Q_N */

#define	CARRY_MASK	0x100000000ULL
#define SIGN_MASK	0x80000000UL
#define U32_MASK	0xffffffffUL
#define	I32_MAX		INT32_MAX
#define I32_MIN		INT32_MIN

/* The following function doesn't require 64-bit compiler support to detect
 * overflow with integer arithmetics. Guess what does the trick. */

int32_t
sat_add_no64bit(int32_t a, int32_t b)
{
	int32_t tmp, res;

	tmp = a + b;

	if (((tmp ^ a) & (tmp ^ b)) < 0) {
		if (tmp & SIGN_MASK)
			res = I32_MAX;
		else
			res = I32_MIN;
	} else {
		res = tmp;
	}

	return res;
}

/* Saturated addition for 32-bit signed integers. */

int32_t
sat_add(int32_t a, int32_t b)
{
	int64_t tmp, res;

	tmp = (int64_t ) a + (int64_t) b;

	if (((CARRY_MASK & tmp) >> 1) != (SIGN_MASK & tmp)) {
		if (SIGN_MASK & tmp)
			res = I32_MAX;
		else
			res = I32_MIN;
	} else {
		res = tmp & U32_MASK; 
	}

	return res;
}

uint32_t
fixmath_add(uint32_t a, uint32_t b)
{
	return a + b;
}

uint32_t
fixmath_mul(uint32_t a, uint32_t b)
{
	uint64_t tmp;

	tmp = (uint64_t) a * (uint64_t) b;

	/* Round up to half. */
	tmp += 1 << (Q_N-1);

	return (tmp >> Q_N);
}

uint32_t
fixmath_div(uint32_t a, uint32_t b)
{
	uint64_t tmp;

	tmp = (uint64_t) a << Q_N;

	/* Round up to half. */
	tmp += b >> 1;

	return tmp / b;
}

int
main(int argc, char *argv[])
{
	double a, b, c;
	uint32_t fa, fb, fc;

	a = 39.332;
	b = 9.890;

	/* Scale and convert floating-point to integer. */
	fa = a * (double) Q_SCALE;
	fb = b * (double) Q_SCALE;

	printf("a=%f b=%f, Q16.16: a=%d b=%d\n", a, b, fa, fb);

	/* Fixed point arithmetics. */

	/* fc = fa + fb */
	fc = fixmath_add(fa, fb);
	c = fc / (double) Q_SCALE;

	printf("a+b: fp=%u c=%f == %f\n", fc, c, a + b);

	fc = fixmath_mul(fa, fb);
	c = fc / (double) Q_SCALE;

	printf("a*b: fp=%u c=%f == %f\n", fc, c, a * b);

	fc = fixmath_div(fa, fb);
	c = fc / (double) Q_SCALE;

	printf("a/b: fp=%u c=%f == %f\n", fc, c, a / b);

	/* Saturated addition. */

	printf("\n");
	printf("Now checking saturated arithmetics...\n");
	printf("\n");

	fa = INT32_MAX;
	fb = INT32_MAX - 124;

	printf("We will try to add %d and %d -- obvious overflow!\n", fa, fb);

	printf("SAT(a+b)=%d NONSAT(a+b)=%d\n", sat_add_no64bit(fa, fb), fa + fb);

	fa = INT32_MIN;
	fb = INT32_MIN + 44;

	printf("We will try to add %d and %d -- obvious overflow too!\n", fa, fb);

	printf("SAT(a+b)=%d NONSAT(a+b)=%d\n", sat_add_no64bit(fa, fb), fa + fb);

	fa = 12844;
	fb = 29244;

	if ((fa + fb) != sat_add_no64bit(fa, fb) ||
	    (fa - fb) != sat_add_no64bit(fa, -fb))
		printf("something is wrong with your saturated math!\n");
	else
		printf("saturated arithmetics seems to be ok...\n");

	return 0;
} 

